/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

/* include values determined at compile time */
#include "clex.h.inc"

/* useful macros */
#define ARRAY_SIZE(X)	(sizeof(X) / sizeof(X[0]))
#define TSET(X)			((X) ? 1 : ((X) = 1, 0))	/* test & set */
#define TCLR(X)			((X) ? ((X) = 0, 1) : 0)	/* test & clear */
#define CH_ESC			'\033'						/* ASCII escape */
#define CH_CTRL(X)		((X) & 0x1F)				/* ASCII ctrl-X */

typedef unsigned short int FLAG;	/* true or false */
typedef short int CODE;				/* usually some #define-d value */

/* operation modes */
/* value 0 is reserved, it means mode unchanged in control_loop() */
#define MODE_COMPLETION				 1
#define MODE_CONFIG_CONFIRM			 2
#define MODE_CONFIG_EDIT_NUM		 3
#define MODE_CONFIG_EDIT_STR		 4
#define MODE_CONFIG_MAIN			 5
#define MODE_DESELECT				 6
#define MODE_DIR_EDIT				 7
#define MODE_DIR_MAIN				 8
#define MODE_DIR_SPLIT				 9
#define MODE_FILE_MAIN				10
#define MODE_HELP					11
#define MODE_HIST					12
#define MODE_LOCATE					13
#define MODE_MENU					14
#define MODE_SELECT					15
#define MODE_SORT					16
/* pseudo-modes */
#define MODE_SPECIAL_QUIT			98
#define MODE_SPECIAL_RETURN			99

/* see ustring.c for details */
typedef struct {
	char *str;			/* space to hold some character string */
	size_t memalloc;	/* size of the allocated memory */
} USTRING;

/* info about screen display */
typedef struct {
	FLAG curses;			/* curses active */
	FLAG wait;				/* a message has been written to the
							   text-mode screen, wait for a keypress
							   before starting curses */
	int scrcols;			/* number of columns */
	int panlines;			/* number of lines in the panel area */
	int textline_area;		/* total room for the textline */
} DISPLAY;

typedef struct {
	const char *login_at_host;	/* my name: user@host */
	const char *shell;			/* my login shell */
	const char *homedir;		/* my home directory */
	FLAG isroot;				/* my effective uid == 0 */
} CLEX_USER;

/* line of text where the user can enter and edit his/her input */
typedef struct {
	const char *prompt;	/* prompt */
	int promptlen;		/* prompt length */
	USTRING line;		/* user's input */
	int size;			/* number of chars in the line */
	int curs;			/* cursor position from 0 to 'size' */
	int offset;			/* offset - when the line is too long,
						   first 'offset' characters are hidden */
} TEXTLINE;

/********************************************************************/

/* panel types */
#define PANEL_TYPE_COMPL		 1
#define PANEL_TYPE_CONFIG		 2
#define PANEL_TYPE_DIR_BROWSE	 3
#define PANEL_TYPE_DIR_SPLIT	 4
#define PANEL_TYPE_FILE			 5
#define PANEL_TYPE_HELP			 6
#define PANEL_TYPE_HIST			 7
#define PANEL_TYPE_LOCATE		 8
#define PANEL_TYPE_MENU			 9
#define PANEL_TYPE_SORT			10
#define PANEL_TYPE_EMPTY		98	/* pseudo type */
#define PANEL_TYPE_NONE			99	/* pseudo type */

/*
 * common part of every PANEL_XXX structure:
 *
 * int cnt, top, curs, min;		panel lines: total count, top of the
 *								screen, cursor, and the first line
 * CODE type;					panel type: one of PANEL_TYPE_XXX
 *
 * first line 'min' is either 0 or -1. If set to -1 a special line
 * numbered -1 "--- (leave this panel) ---" is inserted before
 * real first panel line which is always line number 0.
 *
 * if 'cnt' is zero, panel is empty and remaining panel values
 * are invalid
 */

#define PANEL_COMMON	int cnt, top, curs, min; CODE type;

typedef struct {
	PANEL_COMMON
	/* panel specific data follows here */
} PANEL;

/********************************************************************/

/* file types recognized in the file panel */
#define FT_PLAIN_FILE		 0
#define FT_PLAIN_EXEC		 1
#define FT_PLAIN_SUID		 2
#define FT_PLAIN_SUID_ROOT	 3
#define FT_PLAIN_SGID		 4
#define FT_DIRECTORY		 5
#define FT_DIRECTORY_MNT	 6
#define FT_DEV_BLOCK		 7
#define FT_DEV_CHAR			 8
#define FT_FIFO				 9
#define FT_SOCKET			10
#define FT_OTHER			11
#define FT_NA				12

/* file type tests */
#define IS_FT_PLAIN(X)		((X) >= 0 && (X) <= 4)
#define IS_FT_EXEC(X)		((X) >= 1 && (X) <= 4)
#define IS_FT_DIR(X)		((X) >= 5 && (X) <= 6)
#define IS_FT_DEV(X)		((X) >= 7 && (X) <= 8)

/* text buffer sizes */
#define FE_LINKS_STR	 5		/* e.g. =2__, =999, >999 */
#define FE_TIME_STR		 8		/* e.g. 20jul01, __18:00, _06:00p */
#define FE_SIZE_STR		 9		/* e.g. ____551_, 112.890k */
#define FE_MODE1_STR	 5		/* e.g. 0644 */
#define FE_MODE2_STR	10		/* e.g. rw-r--r-- */
#define FE_OWNER_STR	18		/* e.g. ____root:mail____ */
/* FE_DEV_STR determined at compile time, e.g. _21:0x07 */

/* file sort order */
#define SORT_NAME		0
#define SORT_SUFFIX		1
#define SORT_SIZE		2
#define SORT_SIZE_REV	3
#define SORT_TIME		4
#define SORT_TIME_REV	5
#define SORT_EMAN		6		/* Name <--> emaN */

/*
 * file description - exhausting, isn't it ?
 * we allocate many of these, bitfields save memory
 */
typedef struct {
	USTRING file;			/* file name */
	USTRING link;			/* where the symbolic link points to */
	const char *extension;	/* file name extension (suffix) */

	time_t mtime;			/* last file modification */
	off_t size;				/* file size */
	CODE file_type;			/* one of FT_XXX */
	unsigned int select:1;		/* flag: this entry is selected */
	unsigned int symlink:1;		/* flag: it is a symbolic link */
	unsigned int dotdir:2;		/* . (1) or .. (2) directory */
									/* in readable text format: */
	char atime_str[FE_TIME_STR];	/* access time */
	char ctime_str[FE_TIME_STR];	/* inode change time */
	char links_str[FE_LINKS_STR];	/* number of links (if > 1) */
	char mode1_str[FE_MODE1_STR];	/* permissions - octal number */
	char mode2_str[FE_MODE2_STR];	/* permissions - string */
	char mtime_str[FE_TIME_STR];	/* modification time */
	char owner_str[FE_OWNER_STR];	/* owner and group */
	union {
		char size[FE_SIZE_STR];		/* file size */
		char device[FE_DEV_STR];	/* device info (major,minor) */
	} sd_str;
} FILE_ENTRY;

typedef struct ppanel_file {
	PANEL_COMMON
	USTRING dir;			/* working directory */
	struct ppanel_file *other;	/* primary <--> secondary panel ptr */
	int selected;			/* number of selected entries */
	FLAG expired;			/* expiration: panel needs re-read */
	int fe_slots;			/* allocated slots in 'files' below */
	FILE_ENTRY **files;		/* main part: list of files in panel's
							   working directory 'dir' */
} PANEL_FILE;

/********************************************************************/

typedef struct {
	PANEL_COMMON
	int fe_slots;			/* same meaning as in PANEL_FILE above */
	FILE_ENTRY **files;		/* ditto */
} PANEL_LOCATE;

/********************************************************************/

typedef struct {
	PANEL_COMMON
	CODE order;				/* file sort order: one of SORT_XXX */
	const char **menu;		/* menu of filename sort options */
} PANEL_SORT;

/********************************************************************/

typedef struct {
	PANEL_COMMON
	struct panel_dir_entry {
		const char *name;	/* directory name */
		int shlen;			/* length of the repeating name part */
							/* used for fancy formatting */
	} *dir;					/* list of directories to choose from */
} PANEL_DIR_BROWSE;

typedef struct {
	PANEL_COMMON
	const char *dirname;	/* source directory name */
} PANEL_DIR_SPLIT;

#define TOPNDIRS	5		/* top N recently visited directories */
							/* are treated specially in the panel */

/********************************************************************/

#define CFGVAR_MAXLEN	16
#define CFGVALUE_MAXLEN	60

typedef struct {
	const char var[CFGVAR_MAXLEN + 1];	/* name of the variable */
	const char *help;		/* one line help */
	void *table;			/* ptr to an internal table with details */
	unsigned int isnum:1;	/* is numeric (otherwise string) */
	unsigned int changed:1;	/* value changed */
	unsigned int saveit:1;	/* value should be saved to disk */
} CONFIG_ENTRY;

typedef struct {
	PANEL_COMMON
	CONFIG_ENTRY *config;	/* list of all configuration variables */
} PANEL_CONFIG;

/********************************************************************/

typedef struct {
	PANEL_COMMON
	USTRING **list;			/* list of previously executed commands */
} PANEL_HISTORY;

/********************************************************************/

typedef struct {
	PANEL_COMMON
	CODE pagenum;			/* internal number of current page */
	const char *heading;	/* heading of the current help page */
	const char **txt;		/* help text to be displayed */
} PANEL_HELP;

/********************************************************************/

/*
 * completion types:
 *   any file, directory, executable, command in $PATH, username,
 *   and environment variable
 */
#define COMPL_TYPE_FILE		0
#define COMPL_TYPE_DIR		1
#define COMPL_TYPE_CMD		2
#define COMPL_TYPE_PATHCMD	3
#define COMPL_TYPE_USER		4
#define COMPL_TYPE_ENV		5

typedef struct {
	USTRING name;			/* name suitable for a completion */
	FLAG isdir;				/* it is a directory name */
	const char *aux;		/* additional information */
} CANDIDATE;

typedef struct {
	PANEL_COMMON
	CODE compl_type;		/* one of COMPL_TYPE_XXX */
	CANDIDATE *candidate;	/* menu of completion candidates */
} PANEL_COMPLETION;

/********************************************************************/

typedef struct {
	PANEL_COMMON
	const char **menu;		/* menu to be displayed */
} PANEL_MENU;

/********************************************************************/

/* global variables */

extern DISPLAY display;
extern CLEX_USER clex_user;
extern TEXTLINE *textline;		/* --> active line */
extern TEXTLINE line_cmd, line_dir, line_tmp;
extern PANEL *panel;			/* --> active panel */
extern PANEL_FILE *ppanel_file;
extern PANEL_LOCATE panel_locate;
extern PANEL_DIR_BROWSE panel_dir_browse;	
extern PANEL_DIR_SPLIT panel_dir_split;	
extern PANEL_HISTORY panel_hist;
extern PANEL_CONFIG panel_config;
extern PANEL_SORT panel_sort;
extern PANEL_HELP panel_help;
extern PANEL_COMPLETION panel_compl;
extern PANEL_MENU panel_menu, panel_config_confirm;
extern CODE next_mode;			/* see control.c comments */
extern FLAG adminmode;			/* flag: clex -a */
