/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* pid_t */
#include <errno.h>			/* errno */
#include <signal.h>			/* sigaction() */
#include <stdio.h>			/* puts() */
#include <stdlib.h>			/* exit() */
#include <string.h>			/* strcmp() */
#include <unistd.h>			/* fork() */

/* waitpid() */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include "clex.h"
#include "exec.h"

#include "cfg.h"			/* config_str() */
#include "control.h"		/* get_current_mode() */
#include "edit.h"			/* edit_islong() */
#include "err.h"			/* err_exit() */
#include "inout.h"			/* win_remark() */
#include "history.h"		/* hist_save() */
#include "list.h"			/* list_directory() */
#include "tty.h"			/* tty_setraw() */
#include "util.h"			/* base_name() */
#include "ustring.h"		/* us_copy() */
#include "xterm_title.h"	/* xterm_title_set() */

extern int errno;

#define MAX_ARGS	8
static char *shell_argv[MAX_ARGS + 1];
static int cmd_index;

void
exec_initialize(void)
{
	US_INIT(line_cmd.line);

	exec_shell_reconfig();
	exec_prompt_reconfig();	/* prompt needs to know the shell */
}

static int
parse_shellprog(const char *shellprog)
{
	char ch, *pch;
	static USTRING dup = { 0,0 };
	int i;

	us_copy(&dup,shellprog);
	pch = USTR(dup);

	/* split to arguments */
	for (i = 1, shell_argv[0] = pch; (ch = *pch); pch++ )
		if (ch == ' ') {
			*pch = '\0';
			if (pch[1] != ' ') {
				if (i >= MAX_ARGS) {
					txt_printf("CONFIG: Incorrect \"SHELLPROG\": "
					  "too many arguments\n");
					return -1;
				}
				shell_argv[i++] = pch + 1;
			}
		}
	shell_argv[i] = 0;

	while ( --i >= 1)
		if (strcmp(shell_argv[i],"<COMMAND>") == 0) {
			cmd_index = i;	/* real command belongs here */
			return 0;		/* finished ! */
		}

	txt_printf("CONFIG: Incorrect \"SHELLPROG\": "
	  "missing \"<COMMAND>\"\n");
	return -1;
}

/* prepare to execute commands according to "SHELLPROG" */
void
exec_shell_reconfig(void)
{
	const char *shellprog;

	shellprog = config_str(CFG_SHELLPROG);
	if (*shellprog) {
		if (parse_shellprog(shellprog) == 0)
			return;		/* success */
		txt_printf("EXEC: Will use your login shell \"%s\"\n",
		  clex_user.shell);
	}

	/* no SHELLPROG => automatically set: loginshell -c <COMMAND> */
	((const char **)shell_argv)[0] = clex_user.shell;
	shell_argv[1] = "-c";
	cmd_index = 2;
	shell_argv[3] = 0;
}

void
exec_prompt_reconfig(void)
{
	int len;
	const char *prompt, *chars;
	static USTRING cmd_prompt = { 0,0 };
	static char trailer[4] = " ? ";

	prompt = config_str(CFG_PROMPT);

	if (prompt[0] == '\0') {
		/* autoselect prompt */
		prompt = base_name(shell_argv[0]);
		len = strlen(prompt);
		if (len >= 2 && strcmp(prompt + len - 2,"sh") == 0)
			if (len >= 3 && prompt[len - 3] == 'c')
				chars = "%#";	/* looks like C-shell */
			else
				chars = "$#";	/* looks like Bourne shell */
		else
			chars = ">>";		/* some other shell */
		trailer[1] = chars[clex_user.isroot];
	}
	else
		trailer[1] = '\0';

	us_cat(&cmd_prompt,clex_user.isroot ? "ROOT " : "",
	  prompt,trailer,(char *)0);
	edit_setprompt(&line_cmd,USTR(cmd_prompt));
}

#ifndef HAVE_STRERROR
static char *
strerror(int errcode)
{
	static buffer[32];

	sprintf(buffer,"errno = %d",errcode);
	return buffer;
}
#endif

static void
execute(const char *command)
{
	pid_t childpid;
	int status, code;
	struct sigaction act;

	xterm_title_set(1,command);
	childpid = fork();
	if (childpid == -1)
		printf("EXEC: Cannot create new process (%s)\n",
		  strerror(errno));
	else if (childpid == 0) {
		/* reset signal dispositions */
		act.sa_handler = SIG_DFL;
		act.sa_flags = 0;
		sigemptyset(&act.sa_mask);
		sigaction(SIGINT,&act,0);
		sigaction(SIGQUIT,&act,0);

		/* execute the command */
		((const char **)shell_argv)[cmd_index] = command;
		execv(shell_argv[0],shell_argv);
		printf("EXEC: Cannot execute shell %s (%s)\n",
		  shell_argv[0],strerror(errno));
		exit(99);
		/* NOTREACHED */
	}
	else {
		/* wait for child process exit or stop */
		while (waitpid(childpid,&status,0) < 0)
			/* ignore EINTR */;

		tty_reset();
		if (WIFEXITED(status)) {
			code = WEXITSTATUS(status);
			if (code == 0)
				printf("\nCommand successful. ");
			else
				printf("\nExit code = %d. ",code);
		}
		else {
			code = WTERMSIG(status);
			printf("\nAbnormal termination, signal %d",code);
#ifdef SYS_SIGLIST_DECLARED
			printf(" (%s)",sys_siglist[code]);
#endif
			if (WCOREDUMP(status))
				fputs(", core image dumped",stdout);
			putchar('\n');
		}
	}

	/*
	 * list directory while the user stares at the screen
	 * reading the command output, CLEX appears to restart
	 * very much faster. All output routines called from
	 * list_directory() must check if curses mode is active,
	 * win_warning() is the only function designed for this
	 * situation.
	 */
	xterm_title_set(0,command);
	list_directory();
	ppanel_file->other->expired = 1;

	tty_press_enter(); 
	xterm_title_set(0,0);
}

static int
user_confirm(void)
{
	int in, confirmed;

	fputs("\nExecute the command ? (y = YES) ",stdout);
	fflush(stdout);
	tty_setraw();
	in = getchar();
	tty_reset();
	confirmed = in == 'y' || in == 'Y';
	puts(confirmed ? "yes\n" : "no\n");
	fflush(stdout);
	return confirmed;
}

static int
test_word(const char *line, const char *word, int wordlen)
{
	while (*line == ' ')
		line++;
	return strncmp(line,word,wordlen) == 0;
}

/*
 * return value is the warning level:
 *   0 = no warning issued
 *   1 = only mandatory warning(s)
 *   2 = configurable warning(s)
 */
static int
print_warnings(const char *cmd)
{
	static USTRING cwd = { 0,0 };
	int warn;

	warn = 0;		/* 0 = no warning yet */

	if (get_cwd_us(&cwd) < 0) {
		puts("WARNING: current working directory is not accessible");
		warn = 1;	/* 1 = this warning cannot be turned off */
		us_copy(&cwd,"???");
	}
	else if (strcmp(USTR(ppanel_file->dir),USTR(cwd))) {
		printf("WARNING: current working directory has been renamed:\n"
			   "         old name: %s\n"
			   "         new name: %s\n",
		  USTR(ppanel_file->dir),USTR(cwd));
		us_copy(&ppanel_file->dir,USTR(cwd));
		warn = 1;
	}

	if (config_num(CFG_WARN_RM) && test_word(cmd,"rm ",3)) {
		printf("working directory: %s\n"
			   "WARNING: rm command deletes files, please confirm\n",
			USTR(cwd));
		warn = 2;	/* 2 = this warning can be turned off */
	}

	/* following warnings are not appropriate in history mode */
	if (get_current_mode() == MODE_FILE_MAIN) {
		if (config_num(CFG_WARN_CD) && test_word(cmd,"cd ",3)) {
			puts("WARNING: External cd command will NOT "
			  "change directory in CLEX");
			warn = 2;
		}
		/*
		 * edit_islong() works with 'textline' instead of 'cmd',
		 * because it is meaningfull only in MODE_FILE_MAIN
		 */
		if (config_num(CFG_WARN_LONG) && edit_islong()) {
			puts("WARNING: This long command did not fit on the"
			  " command line");
			warn = 2;
		}
	}

	fflush(stdout);
	return warn;
}

/*
 * function returns 1 if the command has been executed
 * (successfully or not), otherwise 0
 */
int
execute_cmd(const char *cmd)
{
	static FLAG hint = 1;
	int do_exec, warn_level;

	curses_stop();
	putchar('\n');
	puts(cmd);
	putchar('\n');
	fflush(stdout);

	warn_level = print_warnings(cmd);
	do_exec = warn_level == 0 || user_confirm();
	if (do_exec) {
		execute(cmd);
		hist_save(cmd);
	}

	curses_restart();
	if (warn_level >= 2 && TCLR(hint))
		win_remark("warnings can be turned off in the config panel");

	return do_exec;
}
