/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* clex.h */
#include <sys/stat.h>	/* fstat() */
#include <fcntl.h>		/* open() */
#include <stdlib.h>		/* free() */
#include <string.h>		/* strcmp() */
#include <unistd.h>		/* read(), close() */

#include "clex.h"
#include "help.h"

#include "cfg.h"		/* config_str() */
#include "err.h"		/* err_exit() */
#include "control.h"	/* get_previous_mode() */
#include "inout.h"		/* win_panel() */
#include "panel.h"		/* pan_adjust() */
#include "util.h"		/* emalloc() */

#ifndef SSIZE_MAX
# define SSIZE_MAX 32767
#endif

/* limits */
#define HELP_HISTORY_SIZE	16
#define HELP_PAGE_MAXCNT	50			/* pages max */
#define HELP_FILE_MAXSIZE	50000L		/* bytes max */

static const char *builtin_help[] = {
#include "help.c.inc"		/* "line1","line2",....,"lineN", */
0
};

static char *helptext = 0;			/* contents of the helpfile */
static const char **helplines = 0;	/* helpfile split into lines */
static struct {
	const char *name;			/* internal page name */
	const char *heading;		/* heading to be displayed */
	int firstline;				/* first line = 'helplines[]' index */
	int size;					/* total number of lines */
} helppage[HELP_PAGE_MAXCNT];
static int pagecnt;				/* number of pages in 'helppage[]' */
static int tocnum;				/* ToC page number */

static const char *helpfile;	/* name of the helpfile */
static FLAG helperror;			/* there was an error */

/* help history - allows only to go back to previous page */
static struct {
	int pagenum, top, curs;
} history[HELP_HISTORY_SIZE];
static int head, tail;			/* circular buffer indices */

/* increment/decrement (+1/-1) circular buffer index X */
#define MVINDEX(X,N) \
  (X = (X + HELP_HISTORY_SIZE + N) % HELP_HISTORY_SIZE)

static int
page2num(const char *pagename)
{
	int i;

	for (i = 0; i < pagecnt; i++)
		if (strcmp(helppage[i].name,pagename) == 0)
			return i;
	return -1;		/* no such page */
}

static void
help_error_init(const char *filename)
{
	helpfile = filename;
	helperror = 0;
}

static void
help_error(const char *msg)
{
	if (!TSET(helperror))
		txt_printf("HELP: Parsing helpfile \"%s\"\n",helpfile);
	txt_printf("HELP: %s\n",msg);
}

/*
 * helpfile format:
 * note: lines beginning with '#' on the first position
 *       are ignored (comments)
 *
 * helpfile:
 *    @V=VERSION    <-- program version
 *    <page1>
 *    <page2>
 *    ....
 *    <pageX>
 *
 * page:
 *    @P=pgname     <-- new page named 'pgname' starts here
 *    @H=heading    <-- this is optional, default heading = page name
 *    <line1>
 *    <line2>
 *    ....
 *    <lineY>
 *
 * line:
 *    literal text
 *      or
 *    @L=pgname     <-- next line is a link to the page 'pgname'
 */

/*
 * implementation note: @L= lines are merged with the next line:
 * text1			text1
 * @L=link ------>	@L=link<null>text2
 * text2			text3
 * text3   
 */

static int
read_builtin_help(void)
{
	FLAG pagestart;
	char ch;
	const char *ptr;
	int linecnt;

	help_error_init("<built-in help>");
	helplines = builtin_help;

	/* divide to pages */
	pagestart = 0;
	for (linecnt = pagecnt = 0; (ptr = helplines[linecnt]); linecnt++) {
		if (ptr[0] == '@' && (ch = ptr[1]) && ptr[2] == '=') {
			/* process '@X=' directive */
			if (ch == 'V') {
				/* version stamp */
				if (strcmp(ptr + 3,VERSION)) {
					help_error("VERSION mismatch");
					return -1;
				}
				continue;
			} else if (ch == 'P') {
				/* page begin */
				if (pagecnt == HELP_PAGE_MAXCNT) {
					help_error("Too many help pages");
					return -1;
				}
				pagestart = 1;
				helppage[pagecnt].name = ptr + 3;
				helppage[pagecnt].heading = ptr + 3; /* @H= overrides */
				helppage[pagecnt].size = 0;
				pagecnt++;
				continue;
			} else if (ch == 'H') {
				/* header */
				if (!pagestart) {
					help_error("@H not after @P");
					return -1;
				}
				helppage[pagecnt - 1].heading = ptr + 3;
				continue;
			}
			/*
			 * @L= lines are already preprocessed
			 * and can be treated like normal text
			 */
		}
		if (pagecnt == 0) {
			help_error("text before first page");
			return -1;
		}
		if (TCLR(pagestart))
			helppage[pagecnt - 1].firstline = linecnt;
		helppage[pagecnt - 1].size++;
	}

	/* ToC must exist */
	if ( (tocnum = page2num("ToC")) < 0) {
		help_error("Table of contents is missing");
		return -1;
	}

	return 0;
}

/* read external helpfile into memory and parse it */
static int
read_helpfile(const char *helpfile)
{
	int fd, rd, linecnt;
	FLAG mergelink;
	char ch, *ptr;
	struct stat stbuf;
	off_t filesize, left;

	help_error_init(helpfile);
	if ( (fd = open(helpfile,O_RDONLY)) < 0) {
		help_error("Cannot open file");
		return -1;
	}

	/* check file size */
	fstat(fd,&stbuf);	/* cannot fail with valid descriptor */
	filesize = (int)stbuf.st_size;
	if (filesize == 0) {
		close(fd);
		help_error("Helpfile is empty");
		return -1;
	}
	if (filesize > HELP_FILE_MAXSIZE) {
		close(fd);
		help_error("Helpfile is too big");
		return -1;
	}

	/* read into memory */
	ptr = helptext = emalloc(filesize + 1);
	left = filesize;
	while (left > 0) {
		rd = read(fd,ptr,left > SSIZE_MAX ? SSIZE_MAX : left);
		if (rd <= 0)
			break;
		ptr += rd;
		left -= rd;
	}
	close(fd);
	if (rd < 0 || left > 0) {
		help_error("File read error");
		return -1;
	}
	if (helptext[filesize - 1] != '\n')
		helptext[filesize++] = '\n';

	/* count lines */
	for (linecnt = 0, ptr = helptext; ptr < helptext + filesize; ptr++)
		if (*ptr == '\t' || *ptr == '\r')
			*ptr = ' ';
		else if (*ptr == '\n' || *ptr == '\0') {
			*ptr = '\0';
			linecnt++;
	}
	helplines = emalloc(linecnt * sizeof(char *));

	/* split to lines and divide to pages */
	linecnt = pagecnt = 0;
	for (ptr = helptext; ptr < helptext + filesize; ) {
		if (ptr[0] == '@' && (ch = ptr[1]) && ptr[2] == '=') {
			/* process '@X=' directive */
			if (ch == 'V') {
				/* version stamp */
				if (strcmp(ptr + 3,VERSION))
					help_error("Program and helpfile versions do not"
					  " match; information\n"
					  "      in the on-line help might be inaccurate"
					  " or out-of-date.\n"
					  "      Please consider using the built-in help");
			} else if (ch == 'P') {
				/* page begin */
				mergelink = 0;	/* cannot merge lines across pages */
				if (pagecnt == HELP_PAGE_MAXCNT) {
					help_error("Too many help pages");
					break;
				}
				helppage[pagecnt].name = ptr + 3;
				helppage[pagecnt].heading = ptr + 3; /* @H= overrides */
				helppage[pagecnt].firstline = linecnt;
				helppage[pagecnt].size = 0;
				pagecnt++;
			} else if (pagecnt > 0) {
				if (ch == 'H')
					helppage[pagecnt - 1].heading = ptr + 3;
				else if (ch == 'L') {
					mergelink = 1;
					helplines[linecnt++] = ptr;
					/* do not increment the line counter yet */
				}
			}
		}
		else if (pagecnt == 0 || *ptr == '#')
			/* ignore bogus line or comment */ ;
		else {
			if (!TCLR(mergelink))
				helplines[linecnt++] = ptr;
			helppage[pagecnt - 1].size++;
		}

		/* advance to the next line */
		while (*ptr++)
			;
	}

	/* ToC must exist */
	if ( (tocnum = page2num("ToC")) < 0) {
		help_error("Table of contents is missing");
		return -1;
	}

	return 0;
}

void
help_reconfig(void)
{
	const char *helpfile;

	/* erase previous help */
	if (helptext) {
		free(helptext);
		helptext = 0;
	}
	if (helplines && helplines != builtin_help) {
		free(helplines);
		helplines = 0;
	}

	helpfile = config_str(CFG_HELPFILE);
	if (*helpfile) {
	   if (read_helpfile(helpfile) == 0)
			return;
	   help_error("Using the built-in help instead");
	}
	if (read_builtin_help() < 0)
		err_exit("BUG: built-in help is incorrect");
}

static void
set_page(int pg)
{
	panel_help.top = panel_help.curs = 0;
	panel_help.cnt = helppage[pg].size;
	panel_help.pagenum = pg;
	panel_help.heading = helppage[pg].heading;
	panel_help.txt = helplines + helppage[pg].firstline;
	win_heading();
}

static void
help_goto(const char *pagename)
{
	int pg;

	if ( (pg = page2num(pagename)) < 0) {
		/* ERROR 404 :-) */
		win_warning("HELP: help-page not found.");
		return;
	}

	/* save current page parameters */
	history[head].pagenum = panel_help.pagenum;
	history[head].top = panel_help.top;
	history[head].curs = panel_help.curs;
	MVINDEX(head,1);
	if (head == tail)
		MVINDEX(tail,1);

	/* new page */
	set_page(pg);

	win_panel();
}

void
help_prepare(void)
{
	int pg;
	const char *name;

	head = tail = 0;

	/* CLEX help is context sensitive */
	switch (get_previous_mode()) {
	case MODE_COMPLETION:
		name = "completion";
		break;
	case MODE_CONFIG_MAIN:
	case MODE_CONFIG_EDIT_NUM:
	case MODE_CONFIG_EDIT_STR:
	case MODE_CONFIG_CONFIRM:
		name = "config";
		break;
	case MODE_DIR_EDIT:
	case MODE_DIR_MAIN:
	case MODE_DIR_SPLIT:
		name = "dir";
		break;
	case MODE_FILE_MAIN:
		name = "file";
		break;
	/* there is no case MODE_HELP: */
	case MODE_HIST:
		name = "history";
		break;
	case MODE_LOCATE:
		name = "locate";
		break;
	case MODE_MENU:
		/*
		 * help with the menu itself or help invoked from the menu ?
		 * refer to tab_shortcut[] in control.c
		 */
		name = panel_menu.curs == 0 ? "ToC" : "menu";
		break;
	case MODE_SELECT:
	case MODE_DESELECT:
		name = "select";
		break;
	case MODE_SORT:
		name = "sort";
		break;
	default:
		name = "ToC";
	}

	if ( (pg = page2num(name)) < 0) {
		win_warning("HELP: help-page not found.");
		pg = tocnum;
	}
	set_page(pg);

	panel = (PANEL *)&panel_help;
	textline = 0;
	win_remark("<F1> = Table of Contents, ctrl-C = leave help");
}

/* follow link */
void
cx_help_link(void)
{
	const char *str;

	str = panel_help.txt[panel_help.curs];
	if (str[0] == '@' && str[1] == 'L' && str[2] == '=')
		help_goto(str + 3);
}

/* display table of contents */
void
cx_help_contents(void)
{
	if (panel_help.pagenum != tocnum)
		help_goto("ToC");
}

/* go back to previous page */
void cx_help_back(void)
{
	if (head == tail) {
		win_remark("there is no previous help-page");
		return;
	}

	MVINDEX(head,-1);
	set_page(history[head].pagenum);
	panel_help.top = history[head].top;
	panel_help.curs = history[head].curs;
	pan_adjust((PANEL *)&panel_help);
	win_panel();
}
