/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>	/* clex.h */
#include <ctype.h>		/* iscntrl() */
#include <signal.h>		/* signal() */
#include <stdarg.h>		/* va_list */
#include <stdio.h>		/* printf() */
#include <stdlib.h>		/* getenv() */
#include <string.h>		/* strchr() */
#include <unistd.h>		/* fork() */

/* waitpid() */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include "clex.h"
#include "xterm_title.h"

#include "cfg.h"		/* config_num() */
#include "util.h"		/* estrdup() */

static FLAG enabled = 0;
static const char *oldtitle = 0;

#define XPROP_TIMEOUT	6	/* timeout for xprop command in seconds */
#define MAX_CMD	20	/* max length of executed command in the title */

void
xterm_title_initialize(void)
{
	xterm_title_reconfig();
	xterm_title_set(0,0);
}

/*
 * run the command
 *   xprop -id $WINDOWID WM_NAME
 * to get the current xterm title
 *
 * errors are not reported, we don't want to confuse the user
 */
static const char
*get_title(void)
{
	int left, rd, fd[2];
	char *wid, *buff, *p1, *p2;
	char title[128];
	pid_t pid;

	if ( (wid = getenv("WINDOWID")) == 0)
		return 0;

	if (pipe(fd) < 0 || (pid = fork()) < 0)
		return 0;

	if (pid == 0) {
		/* this is the child process */
		close(fd[0]);	/* close read end */
		if (fd[1] != STDOUT_FILENO) {
			if (dup2(fd[1], STDOUT_FILENO) != STDOUT_FILENO)
				_exit(126);
			close(fd[1]);
		}
		signal(SIGALRM,SIG_DFL);
		alarm(XPROP_TIMEOUT);
		execlp("xprop", "xprop", "-id", wid, "WM_NAME", (char *)0);
		_exit(127);
	}

	/* parent continues here */
	/* do not return before waitpid() or you create a zombie */
	close(fd[1]);	/* close write end */
	buff = title;
	left = sizeof(title) - 1 /* 1 byte for trailing null */;
	while (left > 0) {
		rd = read(fd[0],buff,left);
		if (rd <= 0)
			break;
		left -= rd;
		buff += rd;
	}
	*buff = '\0';
	close(fd[0]);

	if (waitpid(pid, 0, 0) < 0)
		return 0;

	/* rd < 0 means an error, rd > 0 means buffer too small */
	if (rd)
		return 0;

	/* get the window title in quotation marks */
	p1 = strchr(title,'\"');
	if (p1 == 0)
		return 0;
	p2 = strchr(++p1,'\"');
	if (p2 == 0)
		return 0;
	*p2 = '\0';

	return estrdup(p1);
}

void
xterm_title_reconfig(void)
{
	int xterm_feature;
	char *term;

	enabled = 0;
	xterm_feature = config_num(CFG_XTERM_TITLE);
	/* 0 = disabled */
	if (xterm_feature == 0)
		return;
	/* 1 = automatic -> check $TERM for xterm, kterm or dtterm */
	if (xterm_feature == 1) {
		if ( (term = getenv("TERM")) == 0)
			return;
		if (strncmp(term,"xterm",5) && strncmp(term,"kterm",5) &&
		  strncmp(term,"dtterm",6))
			return;
	}
	/* otherwise 2 = enabled */
	enabled = 1;
	if (oldtitle == 0 && (oldtitle = get_title()) == 0)
		oldtitle = "terminal";
}

static void
set_xtitle(const char *str1, ...)
{
	va_list argptr;
	const char *strn;

	fputs("\033]0;",stdout);

	fputs(str1,stdout);
	va_start(argptr,str1);
	while ( (strn = va_arg(argptr, char *)) )
		fputs(strn,stdout);
	va_end(argptr);

	/* Xterm FAQ claims \007 is incorrect, but everybody uses it */
	fputs("\007",stdout);
	fflush(stdout);
}

void
xterm_title_set(int busy, const char *cmd)
{
	char *clex, short_cmd[MAX_CMD + 4];	/* 4 chars for "..." + null */
	int i, ch;

	if (!enabled)
		return;

	clex = busy ? "CLEX: " : "clex: ";

	if (cmd == 0)
		set_xtitle(clex,clex_user.login_at_host, (char *)0);
	else {
		/* CLEX is executing 'cmd' */
		for (i = 0; (ch = cmd[i] & 0xFF); i++) {
			if (i == MAX_CMD) {
				short_cmd[i++] = '.';
				short_cmd[i++] = '.';
				short_cmd[i++] = '.';
				break;
			}
			short_cmd[i] = iscntrl(ch) ? '?' : ch;
		}
		short_cmd[i] = '\0';
		set_xtitle(clex,short_cmd,(char *)0);
	}
}

void
xterm_title_restore(void)
{
	if (enabled)
		set_xtitle(oldtitle,(char *)0);
}
